@@ -22,18 +22,18 @@
 
 #include <config.h>
 
-#include <libtelepathy/tp-helpers.h>
 #include <libtelepathy/tp-chan-type-text-gen.h>
 #include <libtelepathy/tp-chan-iface-chat-state-gen.h>
 #include <libtelepathy/tp-conn.h>
+#include <libtelepathy/tp-helpers.h>
 
 #include "empathy-tp-chat.h"
 #include "empathy-contact-manager.h"
 #include "empathy-contact-list.h"
-#include "empathy-session.h"
 #include "empathy-marshal.h"
 #include "gossip-debug.h"
 #include "gossip-time.h"
+#include "gossip-utils.h"
 
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
 		       EMPATHY_TYPE_TP_CHAT, EmpathyTpChatPriv))
@@ -41,43 +41,46 @@
 #define DEBUG_DOMAIN "TpChat"
 
 struct _EmpathyTpChatPriv {
-	EmpathyContactList *list;
-	McAccount          *account;
-	gchar              *id;
-	MissionControl     *mc;
+	EmpathyContactList    *list;
+	EmpathyContactManager *manager;
+	McAccount             *account;
+	gchar                 *id;
+	MissionControl        *mc;
 
-	TpChan             *tp_chan;
-	DBusGProxy         *text_iface;
-	DBusGProxy	   *chat_state_iface;
+	TpChan                *tp_chan;
+	DBusGProxy            *text_iface;
+	DBusGProxy	      *chat_state_iface;
 };
 
-static void empathy_tp_chat_class_init (EmpathyTpChatClass *klass);
-static void empathy_tp_chat_init       (EmpathyTpChat      *chat);
-static void tp_chat_finalize           (GObject            *object);
-static void tp_chat_destroy_cb         (TpChan             *text_chan,
-					EmpathyTpChat      *chat);
-static void tp_chat_received_cb        (DBusGProxy         *text_iface,
-					guint               message_id,
-					guint               timestamp,
-					guint               from_handle,
-					guint               message_type,
-					guint               message_flags,
-					gchar              *message_body,
-					EmpathyTpChat      *chat);
-static void tp_chat_sent_cb            (DBusGProxy         *text_iface,
-					guint               timestamp,
-					guint               message_type,
-					gchar              *message_body,
-					EmpathyTpChat      *chat);
-static void tp_chat_state_changed_cb   (DBusGProxy         *chat_state_iface,
-					guint               handle,
-					EmpathyTpChatState  state,
-					EmpathyTpChat      *chat);
-static void tp_chat_emit_message       (EmpathyTpChat      *chat,
-					guint               type,
-					guint               timestamp,
-					guint               from_handle,
-					const gchar        *message_body);
+static void empathy_tp_chat_class_init (EmpathyTpChatClass        *klass);
+static void empathy_tp_chat_init       (EmpathyTpChat             *chat);
+static void tp_chat_finalize           (GObject                   *object);
+static void tp_chat_destroy_cb         (TpChan                    *text_chan,
+					EmpathyTpChat             *chat);
+static void tp_chat_closed_cb          (TpChan                    *text_chan,
+					EmpathyTpChat             *chat);
+static void tp_chat_received_cb        (DBusGProxy                *text_iface,
+					guint                      message_id,
+					guint                      timestamp,
+					guint                      from_handle,
+					guint                      message_type,
+					guint                      message_flags,
+					gchar                     *message_body,
+					EmpathyTpChat             *chat);
+static void tp_chat_sent_cb            (DBusGProxy                *text_iface,
+					guint                      timestamp,
+					guint                      message_type,
+					gchar                     *message_body,
+					EmpathyTpChat             *chat);
+static void tp_chat_state_changed_cb   (DBusGProxy                *chat_state_iface,
+					guint                      handle,
+					TelepathyChannelChatState  state,
+					EmpathyTpChat             *chat);
+static void tp_chat_emit_message       (EmpathyTpChat             *chat,
+					guint                      type,
+					guint                      timestamp,
+					guint                      from_handle,
+					const gchar               *message_body);
 
 enum {
 	MESSAGE_RECEIVED,
@@ -162,6 +165,9 @@ tp_chat_finalize (GObject *object)
 		g_object_unref (priv->tp_chan);
 	}
 
+	if (priv->manager) {
+		g_object_unref (priv->manager);
+	}
 	if (priv->list) {
 		g_object_unref (priv->list);
 	}
@@ -182,7 +188,6 @@ empathy_tp_chat_new (McAccount *account,
 {
 	EmpathyTpChatPriv     *priv;
 	EmpathyTpChat         *chat;
-	EmpathyContactManager *manager;
 
 	g_return_val_if_fail (MC_IS_ACCOUNT (account), NULL);
 	g_return_val_if_fail (TELEPATHY_IS_CHAN (tp_chan), NULL);
@@ -190,11 +195,11 @@ empathy_tp_chat_new (McAccount *account,
 	chat = g_object_new (EMPATHY_TYPE_TP_CHAT, NULL);
 	priv = GET_PRIV (chat);
 
-	manager = empathy_session_get_contact_manager ();
-	priv->list = empathy_contact_manager_get_list (manager, account);
+	priv->manager = empathy_contact_manager_new ();
+	priv->list = empathy_contact_manager_get_list (priv->manager, account);
 	priv->tp_chan = g_object_ref (tp_chan);
 	priv->account = g_object_ref (account);
-	priv->mc = mission_control_new (tp_get_bus ());
+	priv->mc = gossip_mission_control_new ();
 	g_object_ref (priv->list);
 
 	priv->text_iface = tp_chan_get_interface (tp_chan,
@@ -205,6 +210,9 @@ empathy_tp_chat_new (McAccount *account,
 	g_signal_connect (priv->tp_chan, "destroy",
 			  G_CALLBACK (tp_chat_destroy_cb),
 			  chat);
+	dbus_g_proxy_connect_signal (DBUS_G_PROXY (priv->tp_chan), "Closed",
+				     G_CALLBACK (tp_chat_closed_cb),
+				     chat, NULL);
 	dbus_g_proxy_connect_signal (priv->text_iface, "Received",
 				     G_CALLBACK (tp_chat_received_cb),
 				     chat, NULL);
@@ -235,7 +243,7 @@ empathy_tp_chat_new_with_contact (GossipContact *contact)
 
 	g_return_val_if_fail (GOSSIP_IS_CONTACT (contact), NULL);
 
-	mc = mission_control_new (tp_get_bus ());
+	mc = gossip_mission_control_new ();
 	account = gossip_contact_get_account (contact);
 
 	if (mission_control_get_connection_status (mc, account, NULL) != 0) {
@@ -352,8 +360,8 @@ empathy_tp_chat_send (EmpathyTpChat *chat,
 }
 
 void
-empathy_tp_chat_send_state (EmpathyTpChat      *chat,
-			    EmpathyTpChatState  state)
+empathy_tp_chat_set_state (EmpathyTpChat             *chat,
+			   TelepathyChannelChatState  state)
 {
 	EmpathyTpChatPriv *priv;
 	GError            *error = NULL;
@@ -378,11 +386,7 @@ empathy_tp_chat_send_state (EmpathyTpChat      *chat,
 const gchar *
 empathy_tp_chat_get_id (EmpathyTpChat *chat)
 {
-	EmpathyTpChatPriv  *priv;
-	TpConn             *tp_conn;
-	GArray             *handles;
-	gchar             **names;
-	GError             *error = NULL;
+	EmpathyTpChatPriv *priv;
 
 	g_return_val_if_fail (EMPATHY_IS_TP_CHAT (chat), NULL);
 
@@ -392,12 +396,33 @@ empathy_tp_chat_get_id (EmpathyTpChat *chat)
 		return priv->id;
 	}
 
-	tp_conn = mission_control_get_connection (priv->mc, priv->account, NULL);
-	handles = g_array_new (FALSE, FALSE, sizeof (guint));
-	g_array_append_val (handles, priv->tp_chan->handle);
+	priv->id = empathy_tp_chat_build_id (priv->account, priv->tp_chan);
 
+	return priv->id;
+}
+
+gchar *
+empathy_tp_chat_build_id (McAccount *account,
+			  TpChan    *tp_chan)
+{
+	MissionControl *mc;
+	TpConn         *tp_conn;
+	GArray         *handles;
+	gchar         **names;
+	gchar          *id;
+	GError         *error = NULL;
+	
+	g_return_val_if_fail (MC_IS_ACCOUNT (account), NULL);
+	g_return_val_if_fail (TELEPATHY_IS_CHAN (tp_chan), NULL);
+
+	mc = gossip_mission_control_new ();
+	tp_conn = mission_control_get_connection (mc, account, NULL);
+	g_object_unref (mc);
+
+	handles = g_array_new (FALSE, FALSE, sizeof (guint));
+	g_array_append_val (handles, tp_chan->handle);
 	if (!tp_conn_inspect_handles (DBUS_G_PROXY (tp_conn),
-				      priv->tp_chan->handle_type,
+				      tp_chan->handle_type,
 				      handles,
 				      &names,
 				      &error)) {
@@ -412,14 +437,14 @@ empathy_tp_chat_get_id (EmpathyTpChat *chat)
 	}
 
 	/* A handle name is unique only for a specific account */
-	priv->id = g_strdup_printf ("%s/%s",
-				    mc_account_get_unique_name (priv->account),
-				    *names);
+	id = g_strdup_printf ("%s/%s",
+			      mc_account_get_unique_name (account),
+			      *names);
 
 	g_strfreev (names);
 	g_object_unref (tp_conn);
 
-	return priv->id;
+	return id;
 }
 
 static void
@@ -430,7 +455,7 @@ tp_chat_destroy_cb (TpChan        *text_chan,
 
 	priv = GET_PRIV (chat);
 
-	gossip_debug (DEBUG_DOMAIN, "Channel destroyed");
+	gossip_debug (DEBUG_DOMAIN, "Channel Closed or CM crashed");
 
 	g_object_unref  (priv->tp_chan);
 	priv->tp_chan = NULL;
@@ -440,6 +465,22 @@ tp_chat_destroy_cb (TpChan        *text_chan,
 	g_signal_emit (chat, signals[DESTROY], 0);
 }
 
+static void
+tp_chat_closed_cb (TpChan        *text_chan,
+		   EmpathyTpChat *chat)
+{
+	EmpathyTpChatPriv *priv;
+
+	priv = GET_PRIV (chat);
+
+	/* The channel is closed, do just like if the proxy was destroyed */
+	g_signal_handlers_disconnect_by_func (priv->tp_chan,
+					      tp_chat_destroy_cb,
+					      chat);
+	tp_chat_destroy_cb (text_chan, chat);
+
+}
+
 static void
 tp_chat_received_cb (DBusGProxy    *text_iface,
 		     guint          message_id,
@@ -487,10 +528,10 @@ tp_chat_sent_cb (DBusGProxy    *text_iface,
 }
 
 static void
-tp_chat_state_changed_cb (DBusGProxy         *chat_state_iface,
-			  guint               handle,
-			  EmpathyTpChatState  state,
-			  EmpathyTpChat      *chat)
+tp_chat_state_changed_cb (DBusGProxy                *chat_state_iface,
+			  guint                      handle,
+			  TelepathyChannelChatState  state,
+			  EmpathyTpChat             *chat)
 {
 	EmpathyTpChatPriv *priv;
 	GossipContact     *contact;
@@ -499,6 +540,11 @@ tp_chat_state_changed_cb (DBusGProxy         *chat_state_iface,
 
 	contact = empathy_contact_list_get_from_handle (priv->list, handle);
 
+	gossip_debug (DEBUG_DOMAIN, "Chat state changed for %s (%d): %d",
+		      gossip_contact_get_name (contact),
+		      handle,
+		      state);
+
 	g_signal_emit (chat, signals[CHAT_STATE_CHANGED], 0, contact, state);
 
 	g_object_unref (contact);
